@@ -21,13 +21,14 @@
 #
 
 """
-CPython/cffi wrapper for OpenLDAP's "Lightning" MDB database.
+CPython/CFFI wrapper for OpenLDAP's "Lightning" MDB database.
 
 Please see http://lmdb.readthedocs.org/
 """
 
 from __future__ import absolute_import
 
+import inspect
 import os
 import sys
 import threading
@@ -54,9 +55,9 @@ __all__ += ['Error', 'KeyExistsError', 'NotFoundError', 'PageNotFoundError',
 # Handle moronic Python >=3.0 <3.3.
 UnicodeType = type('')
 if UnicodeType is bytes:
-    UnicodeType = str
+    UnicodeType = unicode
 
-# Used to track context across cffi callbcks.
+# Used to track context across CFFI callbcks.
 _callbacks = threading.local()
 
 _CFFI_CDEF = '''
@@ -129,6 +130,7 @@ _CFFI_CDEF = '''
     int mdb_env_copyfd(MDB_env *env, int fd);
     int mdb_env_stat(MDB_env *env, MDB_stat *stat);
     int mdb_env_info(MDB_env *env, MDB_envinfo *stat);
+    int mdb_env_get_maxkeysize(MDB_env *env);
     int mdb_env_sync(MDB_env *env, int force);
     void mdb_env_close(MDB_env *env);
     int mdb_env_set_flags(MDB_env *env, unsigned int flags, int onoff);
@@ -156,6 +158,7 @@ _CFFI_CDEF = '''
     typedef int (MDB_msg_func)(const char *msg, void *ctx);
     int mdb_reader_list(MDB_env *env, MDB_msg_func *func, void *ctx);
     int mdb_reader_check(MDB_env *env, int *dead);
+    int mdb_dbi_flags(MDB_txn *txn, MDB_dbi dbi, unsigned int *flags);
 
     #define MDB_VERSION_MAJOR ...
     #define MDB_VERSION_MINOR ...
@@ -166,29 +169,45 @@ _CFFI_CDEF = '''
     #define EINVAL ...
     #define ENOMEM ...
     #define ENOSPC ...
+
+    #define MDB_BAD_RSLOT ...
+    #define MDB_BAD_TXN ...
+    #define MDB_BAD_VALSIZE ...
+    #define MDB_CORRUPTED ...
+    #define MDB_CURSOR_FULL ...
+    #define MDB_DBS_FULL ...
+    #define MDB_INCOMPATIBLE ...
+    #define MDB_INVALID ...
+    #define MDB_KEYEXIST ...
+    #define MDB_MAP_FULL ...
+    #define MDB_MAP_RESIZED ...
+    #define MDB_NOTFOUND ...
+    #define MDB_PAGE_FULL ...
+    #define MDB_PAGE_NOTFOUND ...
+    #define MDB_PANIC ...
+    #define MDB_READERS_FULL ...
+    #define MDB_TLS_FULL ...
+    #define MDB_TXN_FULL ...
+    #define MDB_VERSION_MISMATCH ...
+
     #define MDB_APPEND ...
     #define MDB_CREATE ...
-    #define MDB_DBS_FULL ...
     #define MDB_DUPSORT ...
-    #define MDB_KEYEXIST ...
     #define MDB_MAPASYNC ...
-    #define MDB_MAP_FULL ...
     #define MDB_NODUPDATA ...
     #define MDB_NOMETASYNC ...
     #define MDB_NOOVERWRITE ...
+    #define MDB_NORDAHEAD ...
     #define MDB_NOSUBDIR ...
     #define MDB_NOSYNC ...
-    #define MDB_NOTFOUND ...
-    #define MDB_RDONLY ...
-    #define MDB_READERS_FULL ...
-    #define MDB_REVERSEKEY ...
-    #define MDB_TXN_FULL ...
-    #define MDB_NORDAHEAD ...
-    #define MDB_WRITEMAP ...
     #define MDB_NOTLS ...
+    #define MDB_RDONLY ...
+    #define MDB_REVERSEKEY ...
+    #define MDB_WRITEMAP ...
+    #define MDB_NOMEMINIT ...
 
     // Helpers below inline MDB_vals. Avoids key alloc/dup on CPython, where
-    // cffi will use PyString_AS_STRING when passed as an argument.
+    // CFFI will use PyString_AS_STRING when passed as an argument.
     static int pymdb_del(MDB_txn *txn, MDB_dbi dbi,
                          char *key_s, size_t keylen,
                          char *val_s, size_t vallen);
@@ -212,7 +231,7 @@ _CFFI_VERIFY = '''
     #include "lmdb.h"
 
     // Helpers below inline MDB_vals. Avoids key alloc/dup on CPython, where
-    // cffi will use PyString_AS_STRING when passed as an argument.
+    // CFFI will use PyString_AS_STRING when passed as an argument.
     static int pymdb_get(MDB_txn *txn, MDB_dbi dbi, char *key_s, size_t keylen,
                          MDB_val *val_out)
     {
@@ -265,7 +284,7 @@ _CFFI_VERIFY = '''
 if not lmdb._reading_docs():
     import cffi
 
-    # Try to use distutils-bundled cffi configuration to avoid a recompile and
+    # Try to use distutils-bundled CFFI configuration to avoid a recompile and
     # potential compile errors during first module import.
     _config_vars = _config.CONFIG if _config else {
         'extra_compile_args': ['-w'],
@@ -417,12 +436,12 @@ class DiskError(Error):
     MDB_NAME = 'ENOSPC'
 
 # Prepare _error_map, a mapping of integer MDB_ERROR_CODE to exception class.
-_error_map = {}
-for obj in list(globals().values()):
-    if getattr(obj, '__name__', '').endswith('Error'):
-        code = globals().get(getattr(obj, 'MDB_NAME', None))
-        _error_map[code] = obj
-del obj, code
+if not lmdb._reading_docs():
+    _error_map = {}
+    for obj in list(globals().values()):
+        if inspect.isclass(obj) and issubclass(obj, Error) and obj is not Error:
+            _error_map[globals()[obj.MDB_NAME]] = obj
+    del obj
 
 def _error(what, rc):
     """Lookup and instantiate the correct exception class for the error code
@@ -440,7 +459,7 @@ def _kill_dependents(parent):
             child._invalidate()
 
 class Some_LMDB_Resource_That_Was_Deleted_Or_Closed(object):
-    """We need this because cffi on PyPy treats None as cffi.NULL, instead of
+    """We need this because CFFI on PyPy treats None as cffi.NULL, instead of
     throwing an exception it feeds LMDB null pointers. That means simply
     replacing native handles with None during _kill_dependents() will lead to
     NULL pointer derefences. Instead we use this class, and its weird name to
@@ -470,7 +489,7 @@ def _undepend(parent, child):
     parent._deps.pop(id(child), None)
 
 def _mvbuf(mv):
-    """Convert a MDB_val cdata to a cffi buffer object."""
+    """Convert a MDB_val cdata to a CFFI buffer object."""
     return _ffi.buffer(mv.mv_data, mv.mv_size)
 
 def _mvstr(mv):
@@ -480,7 +499,7 @@ def _mvstr(mv):
 def enable_drop_gil():
     """
     Arrange for the global interpreter lock to be released during database IO.
-    This flag is ignored and always assumed to be ``True`` on cffi. Note this
+    This flag is ignored and always assumed to be ``True`` on CFFI. Note this
     can only be set once per process.
 
     Continually dropping and reacquiring the GIL may incur unnecessary overhead
@@ -508,8 +527,7 @@ class Environment(object):
 
     To write to the environment a :py:class:`Transaction` must be created. One
     simultaneous write transaction is allowed, however there is no limit on the
-    number of read transactions even when a write transaction exists. Due to
-    this, write transactions should be kept as short as possible.
+    number of read transactions even when a write transaction exists.
 
     Equivalent to `mdb_env_open()
     <http://symas.com/mdb/doc/group__mdb.html#ga1fe2740e25b1689dc412e7b9faadba1b>`_
@@ -545,7 +563,7 @@ class Environment(object):
             will flush at its discretion, or user can flush with
             :py:meth:`sync`.
 
-        `sync`
+        `sync`:
             If ``False``, never explicitly flush data pages to disk. OS will
             flush at its discretion, or user can flush with :py:meth:`sync`.
             This optimization means a system crash can corrupt the database or
@@ -566,6 +584,15 @@ class Environment(object):
             If ``True`` LMDB will use a writeable memory map to update the
             database. This option is incompatible with nested transactions.
 
+        `meminit`:
+            If ``False`` LMDB will not zero-initialize buffers prior to writing
+            them to disk. This improves performance but may cause old heap data
+            to be written saved in the unused portion of the buffer. Do not use
+            this option if your application manipulates confidential data (e.g.
+            plaintext passwords) in memory. This option is only meaningful when
+            `writemap=False`; new pages are always zero-initialized when
+            `writemap=True`.
+
         `map_async`:
              When ``writemap=True``, use asynchronous flushes to disk. As with
              ``sync=False``, a system crash can then corrupt the database or
@@ -590,26 +617,26 @@ class Environment(object):
             :py:meth:`cursor`. Should match the process's maximum expected
             concurrent transactions (e.g. thread count).
 
-            *Note:* ignored on cffi.
+            *Note:* ignored on CFFI.
 
         `max_spare_cursors`:
             Read-only cursors to cache after becoming unused. Caching cursors
             avoids two allocations per :py:class:`Cursor` or :py:meth:`cursor`
             or :py:meth:`Transaction.cursor` invocation.
 
-            *Note:* ignored on cffi.
+            *Note:* ignored on CFFI.
 
         `max_spare_iters`:
             Iterators to cache after becoming unused. Caching iterators avoids
             one allocation per :py:class:`Cursor` ``iter*`` method invocation.
 
-            *Note:* ignored on cffi.
+            *Note:* ignored on CFFI.
     """
     def __init__(self, path, map_size=10485760, subdir=True,
             readonly=False, metasync=True, sync=True, map_async=False,
-            mode=0o644, create=True, readahead=True, writemap=False,
-            max_readers=126, max_dbs=0, max_spare_txns=1, max_spare_cursors=32,
-            max_spare_iters=32):
+            mode=0o755, create=True, readahead=True, writemap=False,
+            meminit=True, max_readers=126, max_dbs=0, max_spare_txns=1,
+            max_spare_cursors=32, max_spare_iters=32):
         envpp = _ffi.new('MDB_env **')
 
         rc = mdb_env_create(envpp)
@@ -630,8 +657,8 @@ class Environment(object):
         if rc:
             raise _error("mdb_env_set_maxdbs", rc)
 
-        if create and subdir and not os.path.exists(path):
-            os.mkdir(path)
+        if create and subdir and not (os.path.exists(path) or readonly):
+            os.mkdir(path, mode)
 
         flags = MDB_NOTLS
         if not subdir:
@@ -649,11 +676,13 @@ class Environment(object):
             flags |= MDB_NORDAHEAD
         if writemap:
             flags |= MDB_WRITEMAP
+        if not meminit:
+            flags |= MDB_NOMEMINIT
 
         if isinstance(path, UnicodeType):
             path = path.encode(sys.getfilesystemencoding())
 
-        rc = mdb_env_open(self._env, path, flags, mode)
+        rc = mdb_env_open(self._env, path, flags, mode & ~0o111)
         if rc:
             raise _error(path, rc)
         with self.begin(db=object()) as txn:
@@ -662,7 +691,7 @@ class Environment(object):
 
     def close(self):
         """Close the environment, invalidating any open iterators, cursors, and
-        transactions.
+        transactions. Repeat calls to :py:meth:`close` have no effect.
 
         Equivalent to `mdb_env_close()
         <http://symas.com/mdb/doc/group__mdb.html#ga4366c43ada8874588b6a62fbda2d1e95>`_
@@ -686,7 +715,7 @@ class Environment(object):
         rc = mdb_env_get_path(self._env, path)
         if rc:
             raise _error("mdb_env_get_path", rc)
-        return _ffi.string(path[0])
+        return _ffi.string(path[0]).decode(sys.getfilesystemencoding())
 
     def copy(self, path):
         """Make a consistent copy of the environment in the given destination
@@ -695,7 +724,7 @@ class Environment(object):
         Equivalent to `mdb_env_copy()
         <http://symas.com/mdb/doc/group__mdb.html#ga5d51d6130325f7353db0955dbedbc378>`_
         """
-        rc = mdb_env_copy(self._env, path)
+        rc = mdb_env_copy(self._env, path.encode(sys.getfilesystemencoding()))
         if rc:
             raise _error("mdb_env_copy", rc)
 
@@ -774,17 +803,17 @@ class Environment(object):
         """Return some nice environment information as a dict:
 
         +--------------------+---------------------------------------+
-        | map_addr           | Address of database map in RAM.       |
+        | ``map_addr``       | Address of database map in RAM.       |
         +--------------------+---------------------------------------+
-        | map_size           | Size of database map in RAM.          |
+        | ``map_size``       | Size of database map in RAM.          |
         +--------------------+---------------------------------------+
-        | last_pgno          | ID of last used page.                 |
+        | ``last_pgno``      | ID of last used page.                 |
         +--------------------+---------------------------------------+
-        | last_txnid         | ID of last committed transaction.     |
+        | ``last_txnid``     | ID of last committed transaction.     |
         +--------------------+---------------------------------------+
-        | max_readers        | Maximum number of threads.            |
+        | ``max_readers``    | Maximum number of threads.            |
         +--------------------+---------------------------------------+
-        | num_readers        | Number of threads in use.             |
+        | ``num_readers``    | Number of threads in use.             |
         +--------------------+---------------------------------------+
 
         Equivalent to `mdb_env_info()
@@ -795,6 +824,7 @@ class Environment(object):
         if rc:
             raise _error("mdb_env_info", rc)
         return {
+            "map_addr": int(_ffi.cast('long', info.me_mapaddr)),
             "map_size": info.me_mapsize,
             "last_pgno": info.me_last_pgno,
             "last_txnid": info.me_last_txnid,
@@ -802,10 +832,44 @@ class Environment(object):
             "num_readers": info.me_numreaders
         }
 
+    def flags(self):
+        """Return a dict describing Environment constructor flags used to
+        instantiate this environment."""
+        flags_ = _ffi.new('unsigned int[]', 1)
+        rc = mdb_env_get_flags(self._env, flags_)
+        if rc:
+            raise _error("mdb_env_get_flags", rc)
+        flags = flags_[0]
+        return {
+            'subdir': not (flags & MDB_NOSUBDIR),
+            'readonly': bool(flags & MDB_RDONLY),
+            'metasync': not (flags & MDB_NOMETASYNC),
+            'sync': not (flags & MDB_NOSYNC),
+            'map_async': bool(flags & MDB_MAPASYNC),
+            'readahead': not (flags & MDB_NORDAHEAD),
+            'writemap': bool(flags & MDB_WRITEMAP),
+            'meminit': not (flags & MDB_NOMEMINIT)
+        }
+
+    def max_key_size(self):
+        """Return the maximum size in bytes of a record's key part. This
+        matches the ``MDB_MAXKEYSIZE`` constant set at compile time."""
+        return mdb_env_get_maxkeysize(self._env)
+
+    def max_readers(self):
+        """Return the maximum number of readers specified during open of the
+        environment by the first process. This is the same as `max_readers=`
+        specified to the constructor if this process was the first to open the
+        environment."""
+        readers_ = _ffi.new('unsigned int[]', 1)
+        rc = mdb_env_get_maxreaders(self._env, readers_)
+        if rc:
+            raise _error("mdb_env_get_maxreaders", rc)
+        return readers_[0]
+
     def readers(self):
-        """Return a list of newline-terminated human readable strings
-        describing the current state of the reader lock table.
-        """
+        """Return a multi line Unicode string describing the current state of
+        the reader lock table."""
         _callbacks.msg_func = []
         try:
             rc = mdb_reader_list(self._env, _msg_func, _ffi.NULL)
@@ -826,7 +890,7 @@ class Environment(object):
         return reaped[0]
 
     def open_db(self, name=None, txn=None, reverse_key=False, dupsort=False,
-            create=True):
+                create=True):
         """
         Open a database, returning an opaque handle. Repeat
         :py:meth:`Environment.open_db` calls for the same name will return the
@@ -835,6 +899,24 @@ class Environment(object):
         Equivalent to `mdb_dbi_open()
         <http://symas.com/mdb/doc/group__mdb.html#gac08cad5b096925642ca359a6d6f0562a>`_
 
+        Named databases are implemented by *storing a special descriptor in the
+        main database*. All databases in an environment *share the same file*.
+        Because the descriptor is present in the main database, attempts to
+        create a named database will fail if a key matching the database's name
+        already exists. Furthermore *the key is visible to lookups and
+        enumerations*. If your main database keyspace conflicts with the names
+        you use for named databases, then move the contents of your main
+        database to another named database.
+
+        ::
+
+            >>> env = lmdb.open('/tmp/test', max_dbs=2)
+            >>> with env.begin(write=True) as txn
+            ...     txn.put('somename', 'somedata')
+
+            >>> # Error: database cannot share name of existing key!
+            >>> subdb = env.open_db('somename')
+
         A newly created database will not exist if the transaction that created
         it aborted, nor if another process deleted it. The handle resides in
         the shared environment, it is not owned by the current transaction or
@@ -917,6 +999,19 @@ class _Database(object):
             raise _error("mdb_dbi_open", rc)
         self._dbi = dbipp[0]
 
+    def flags(self, txn):
+        """Return the database's associated flags as a dict of _Database
+        constructor kwargs."""
+        flags_ = _ffi.new('unsigned int[]', 1)
+        rc = mdb_dbi_flags(txn._txn, self._dbi, flags_)
+        if rc:
+            raise _error("mdb_dbi_flags", rc)
+        flags = flags_[0]
+        return {
+            'reverse_key': bool(flags & MDB_REVERSEKEY),
+            'dupsort': bool(flags & MDB_DUPSORT),
+        }
+
     def _invalidate(self):
         pass
 
@@ -965,7 +1060,7 @@ class Transaction(object):
 
         `buffers`:
             If ``True``, indicates :py:func:`buffer` objects should be yielded
-            instead of strings. This setting applies to the
+            instead of bytestrings. This setting applies to the
             :py:class:`Transaction` instance itself and any :py:class:`Cursors
             <Cursor>` created within the transaction.
 
@@ -974,6 +1069,13 @@ class Transaction(object):
             returned buffer objects. The benefit of this facility is diminished
             when using small keys and values.
     """
+
+    # If constructor fails, then __del__ will attempt to access these
+    # attributes.
+    _env = _invalid
+    _txn = _invalid
+    _parent = None
+
     def __init__(self, env, db=None, parent=None, write=False, buffers=False):
         _depend(env, self)
         self.env = env # hold ref
@@ -984,7 +1086,8 @@ class Transaction(object):
         self._to_py = _mvbuf if buffers else _mvstr
         self._deps = {}
         if write and env.readonly:
-            raise _error('Cannot start write transaction with read-only env')
+            msg = 'Cannot start write transaction with read-only env'
+            raise _error(msg, EACCES)
         if write:
             flags = 0
         else:
@@ -1053,15 +1156,17 @@ class Transaction(object):
         Equivalent to `mdb_txn_commit()
         <http://symas.com/mdb/doc/group__mdb.html#ga846fbd6f46105617ac9f4d76476f6597>`_
         """
-        if self._txn:
-            _kill_dependents(self)
-            rc = mdb_txn_commit(self._txn)
-            self._txn = _invalid
-            if rc:
-                raise _error("mdb_txn_commit", rc)
+        _kill_dependents(self)
+        rc = mdb_txn_commit(self._txn)
+        self._txn = _invalid
+        if rc:
+            raise _error("mdb_txn_commit", rc)
 
     def abort(self):
-        """Abort the pending transaction.
+        """Abort the pending transaction. Repeat calls to :py:meth:`abort` have
+        no effect after a previously successful :py:meth:`commit` or
+        :py:meth:`abort`, or after the associated :py:class:`Environment` has
+        been closed.
 
         Equivalent to `mdb_txn_abort()
         <http://symas.com/mdb/doc/group__mdb.html#ga73a5938ae4c3239ee11efa07eb22b882>`_
@@ -1098,10 +1203,10 @@ class Transaction(object):
         <http://symas.com/mdb/doc/group__mdb.html#ga4fa8573d9236d54687c61827ebf8cac0>`_
 
             `key`:
-                String key to store.
+                Bytestring key to store.
 
             `value`:
-                String value to store.
+                Bytestring value to store.
 
             `dupdata`:
                 If ``True`` and database was opened with `dupsort=True`, add
@@ -1166,7 +1271,7 @@ class Transaction(object):
 
             value:
                 If the database was opened with dupsort=True and value is not
-                the empty string, then delete elements matching only this
+                the empty bytestring, then delete elements matching only this
                 `(key, value)` pair, otherwise all values for key are deleted.
 
         Returns True if at least one key was deleted.
@@ -1431,7 +1536,7 @@ class Cursor(object):
         """Seek exactly to `key`, returning ``True`` on success or ``False`` if
         the exact key was not found.
 
-        It is an error to :py:meth:`set_key` the empty string.
+        It is an error to :py:meth:`set_key` the empty bytestring.
 
         Equivalent to `mdb_cursor_get()
         <http://symas.com/mdb/doc/group__mdb.html#ga48df35fb102536b32dfbb801a47b4cb0>`_
@@ -1453,7 +1558,7 @@ class Cursor(object):
         ``True`` on success, or ``False`` to indicate key was past end of
         database.
 
-        Behaves like :py:meth:`first` if `key` is the empty string.
+        Behaves like :py:meth:`first` if `key` is the empty bytestring.
 
         Equivalent to `mdb_cursor_get()
         <http://symas.com/mdb/doc/group__mdb.html#ga48df35fb102536b32dfbb801a47b4cb0>`_ with `MDB_SET_RANGE <http://symas.com/mdb/doc/group__mdb.html#ga1206b2af8b95e7f6b0ef6b28708c9127>`_
@@ -1502,10 +1607,10 @@ class Cursor(object):
         <http://symas.com/mdb/doc/group__mdb.html#ga1f83ccb40011837ff37cc32be01ad91e>`_
 
             `key`:
-                String key to store.
+                Bytestring key to store.
 
             `val`:
-                String value to store.
+                Bytestring value to store.
 
             `dupdata`:
                 If ``True`` and database was opened with `dupsort=True`, add
@@ -1545,10 +1650,10 @@ class Cursor(object):
         `set-and-return-previous` operation.
 
             `key`:
-                String key to store.
+                Bytestring key to store.
 
             `value`:
-                String value to store.
+                Bytestring value to store.
         """
         flags = MDB_NOOVERWRITE
         keylen = len(key)
@@ -1576,7 +1681,7 @@ class Cursor(object):
         `delete-and-return-previous` operation.
 
             `key`:
-                String key to delete.
+                Bytestring key to delete.
         """
         if self._cursor_get_key(MDB_SET_KEY, key):
             old = _mvstr(self._val)
