@@ -93,6 +93,7 @@ enum string_id {
     MAX_SPARE_CURSORS_S,
     MAX_SPARE_ITERS_S,
     MAX_SPARE_TXNS_S,
+    MEMINIT_S,
     METASYNC_S,
     MODE_S,
     NAME_S,
@@ -143,6 +144,7 @@ static const char *strings = (
     "max_spare_cursors\0"
     "max_spare_iters\0"
     "max_spare_txns\0"
+    "meminit\0"
     "metasync\0"
     "mode\0"
     "name\0"
@@ -520,6 +522,18 @@ struct dict_field {
     int offset;
 };
 
+/**
+ * Return a new reference to Py_True if the given argument is true, otherwise
+ * a new reference to Py_False.
+ */
+static PyObject *
+py_bool(int pred)
+{
+    PyObject *obj = pred ? Py_True : Py_False;
+    Py_INCREF(obj);
+    return obj;
+}
+
 /**
  * Convert the structure `o` described by `fields` to a dict and return the new
  * dict.
@@ -701,6 +715,25 @@ type_error(const char *what)
     return NULL;
 }
 
+/**
+ * Convert a PyObject to filesystem bytes. Must call fspath_fini() when done.
+ * Return 0 on success, or set an exception and return -1 on failure.
+ */
+static PyObject *
+get_fspath(PyObject *src)
+{
+    if(PyBytes_CheckExact(src)) {
+        Py_INCREF(src);
+        return src;
+    }
+    if(! PyUnicode_CheckExact(src)) {
+        type_error("Filesystem path must be Unicode or bytes.");
+        return NULL;
+    }
+    return PyUnicode_AsEncodedString(src, Py_FileSystemDefaultEncoding,
+                                     "strict");
+}
+
 
 // ----------------
 // Argument parsing
@@ -739,14 +772,14 @@ parse_ulong(PyObject *obj, uint64_t *l, PyObject *max)
     if(rc == -1) {
         return -1;
     } else if(! rc) {
-        type_error("Integer argument must be >= 0");
+        PyErr_Format(PyExc_OverflowError, "Integer argument must be >= 0");
         return -1;
     }
     rc = PyObject_RichCompareBool(obj, max, Py_LE);
     if(rc == -1) {
         return -1;
     } else if(! rc) {
-        type_error("Integer argument exceeds limit.");
+        PyErr_Format(PyExc_OverflowError, "Integer argument exceeds limit.");
         return -1;
     }
 #if PY_MAJOR_VERSION >= 3
@@ -907,7 +940,7 @@ make_trans(EnvObject *env, DbObject *db, TransObject *parent, int write, int buf
     MDB_txn *parent_txn = NULL;
     if(parent) {
         if(parent->flags & TRANS_RDONLY) {
-            return err_set("Read-only transactions cannot be nested.", 0);
+            return err_set("Read-only transactions cannot be nested.", EINVAL);
         }
         if(! parent->valid) {
             return err_invalid();
@@ -916,7 +949,8 @@ make_trans(EnvObject *env, DbObject *db, TransObject *parent, int write, int buf
     }
 
     if(write && env->readonly) {
-        return err_set("Cannot start write transaction with read-only env", 0);
+        const char *msg = "Cannot start write transaction with read-only env";
+        return err_set(msg, EACCES);
     }
 
     TransObject *self = PyObject_New(TransObject, &PyTransaction_Type);
@@ -1065,6 +1099,43 @@ db_clear(DbObject *self)
     return 0;
 }
 
+/**
+ * _Database.flags()
+ */
+static PyObject *
+db_flags(DbObject *self, PyObject *args, PyObject *kwds)
+{
+    struct db_flags {
+        TransObject *txn;
+    } arg = {NULL};
+
+    static const struct argspec argspec[] = {
+        {ARG_TRANS, TXN_S, OFFSET(db_flags, txn)}
+    };
+
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
+    }
+    if(! arg.txn) {
+        return type_error("'txn' argument required");
+    }
+    if(! arg.txn->valid) {
+        return err_invalid();
+    }
+
+    unsigned int f;
+    int rc = mdb_dbi_flags(arg.txn->txn, self->dbi, &f);
+    if(rc) {
+        err_set("mdb_dbi_flags", rc);
+        return NULL;
+    }
+
+    PyObject *dct = PyDict_New();
+    PyDict_SetItemString(dct, "reverse_key", py_bool(f & MDB_REVERSEKEY));
+    PyDict_SetItemString(dct, "dupsort", py_bool(f & MDB_DUPSORT));
+    return dct;
+}
+
 /**
  * _Database.__del__()
  */
@@ -1075,12 +1146,18 @@ db_dealloc(DbObject *self)
     PyObject_Del(self);
 }
 
+static struct PyMethodDef db_methods[] = {
+    {"flags", (PyCFunction)db_flags, METH_VARARGS|METH_KEYWORDS},
+    {0, 0, 0, 0}
+};
+
 static PyTypeObject PyDatabase_Type = {
     PyObject_HEAD_INIT(NULL)
     .tp_basicsize = sizeof(DbObject),
     .tp_dealloc = (destructor) db_dealloc,
     .tp_clear = (inquiry) db_clear,
     .tp_flags = Py_TPFLAGS_DEFAULT,
+    .tp_methods = db_methods,
     .tp_name = "_Database"
 };
 
@@ -1124,7 +1201,7 @@ static PyObject *
 env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 {
     struct env_new {
-        char *path;
+        PyObject *path;
         size_t map_size;
         int subdir;
         int readonly;
@@ -1135,15 +1212,16 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         int create;
         int readahead;
         int writemap;
+        int meminit;
         int max_readers;
         int max_dbs;
         ssize_t max_spare_txns;
         ssize_t max_spare_cursors;
         ssize_t max_spare_iters;
-    } arg = {NULL, 10485760, 1, 0, 1, 1, 0, 0644, 1, 1, 0, 126, 0, 1, 32, 32};
+    } arg = {NULL, 10485760, 1, 0, 1, 1, 0, 0755, 1, 1, 0, 1, 126, 0, 1, 32, 32};
 
     static const struct argspec argspec[] = {
-        {ARG_STR, PATH_S, OFFSET(env_new, path)},
+        {ARG_OBJ, PATH_S, OFFSET(env_new, path)},
         {ARG_SIZE, MAP_SIZE_S, OFFSET(env_new, map_size)},
         {ARG_BOOL, SUBDIR_S, OFFSET(env_new, subdir)},
         {ARG_BOOL, READONLY_S, OFFSET(env_new, readonly)},
@@ -1154,6 +1232,7 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         {ARG_BOOL, CREATE_S, OFFSET(env_new, create)},
         {ARG_BOOL, READAHEAD_S, OFFSET(env_new, readahead)},
         {ARG_BOOL, WRITEMAP_S, OFFSET(env_new, writemap)},
+        {ARG_INT, MEMINIT_S, OFFSET(env_new, meminit)},
         {ARG_INT, MAX_READERS_S, OFFSET(env_new, max_readers)},
         {ARG_INT, MAX_DBS_S, OFFSET(env_new, max_dbs)},
         {ARG_SIZE, MAX_SPARE_TXNS_S, OFFSET(env_new, max_spare_txns)},
@@ -1169,6 +1248,7 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         return type_error("'path' argument required");
     }
 
+    PyObject *fspath_obj = NULL;
     EnvObject *self = PyObject_New(EnvObject, type);
     if(! self) {
         return NULL;
@@ -1200,13 +1280,18 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         goto fail;
     }
 
-    if(arg.create && arg.subdir) {
+    if(! ((fspath_obj = get_fspath(arg.path)))) {
+        goto fail;
+    }
+    const char *fspath = PyBytes_AS_STRING(fspath_obj);
+
+    if(arg.create && arg.subdir && !arg.readonly) {
         struct stat st;
         errno = 0;
-        stat(arg.path, &st);
+        stat(fspath, &st);
         if(errno == ENOENT) {
-            if(mkdir(arg.path, 0700)) {
-                PyErr_SetFromErrnoWithFilename(PyExc_OSError, arg.path);
+            if(mkdir(fspath, arg.mode)) {
+                PyErr_SetFromErrnoWithFilename(PyExc_OSError, fspath);
                 goto fail;
             }
         }
@@ -1235,23 +1320,30 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     if(arg.writemap) {
         flags |= MDB_WRITEMAP;
     }
+    if(! arg.meminit) {
+        flags |= MDB_NOMEMINIT;
+    }
 
-    DEBUG("mdb_env_open(%p, '%s', %d, %o);", self->env, arg.path, flags, arg.mode)
-    UNLOCKED(rc, mdb_env_open(self->env, arg.path, flags, arg.mode));
+    // Strip +x.
+    int mode = arg.mode & ~0111;
+
+    DEBUG("mdb_env_open(%p, '%s', %d, %o);", self->env, fspath, flags, mode)
+    UNLOCKED(rc, mdb_env_open(self->env, fspath, flags, mode));
     if(rc) {
-        err_set(arg.path, rc);
+        err_set(fspath, rc);
         goto fail;
     }
 
     self->main_db = txn_db_from_name(self, NULL, 0);
     if(self->main_db) {
         self->valid = 1;
-        DEBUG("EnvObject '%s' opened at %p", arg.path, self)
+        DEBUG("EnvObject '%s' opened at %p", fspath, self)
         return (PyObject *) self;
     }
 
 fail:
     DEBUG("initialization failed")
+    Py_CLEAR(fspath_obj);
     if(self) {
         env_dealloc(self);
     }
@@ -1309,11 +1401,11 @@ static PyObject *
 env_copy(EnvObject *self, PyObject *args)
 {
     struct env_copy {
-        char *path;
+        PyObject *path;
     } arg = {NULL};
 
     static const struct argspec argspec[] = {
-        {ARG_STR, PATH_S, OFFSET(env_copy, path)}
+        {ARG_OBJ, PATH_S, OFFSET(env_copy, path)}
     };
 
     if(parse_args(self->valid, SPECSIZE(), argspec, args, NULL, &arg)) {
@@ -1322,8 +1414,14 @@ env_copy(EnvObject *self, PyObject *args)
     if(! arg.path) {
         return type_error("path argument required");
     }
+    PyObject *fspath_obj = get_fspath(arg.path);
+    if(! fspath_obj) {
+        return NULL;
+    }
+
     int rc;
-    UNLOCKED(rc, mdb_env_copy(self->env, arg.path));
+    UNLOCKED(rc, mdb_env_copy(self->env, PyBytes_AS_STRING(fspath_obj)));
+    Py_CLEAR(fspath_obj);
     if(rc) {
         return err_set("mdb_env_copy", rc);
     }
@@ -1388,6 +1486,65 @@ env_info(EnvObject *self)
     return dict_from_fields(&info, fields);
 }
 
+/**
+ * Environment.flags() -> dict
+ */
+static PyObject *
+env_flags(EnvObject *self)
+{
+    if(! self->valid) {
+        return err_invalid();
+    }
+
+    unsigned int flags;
+    int rc = mdb_env_get_flags(self->env, &flags);
+    if(rc) {
+        err_set("mdb_env_get_flags", rc);
+        return NULL;
+    }
+
+    PyObject *dct = PyDict_New();
+    PyDict_SetItemString(dct, "subdir", py_bool(!(flags & MDB_NOSUBDIR)));
+    PyDict_SetItemString(dct, "readonly", py_bool(flags & MDB_RDONLY));
+    PyDict_SetItemString(dct, "metasync", py_bool(!(flags & MDB_NOMETASYNC)));
+    PyDict_SetItemString(dct, "sync", py_bool(!(flags & MDB_NOSYNC)));
+    PyDict_SetItemString(dct, "map_async", py_bool(flags & MDB_MAPASYNC));
+    PyDict_SetItemString(dct, "readahead", py_bool(!(flags & MDB_NORDAHEAD)));
+    PyDict_SetItemString(dct, "writemap", py_bool(flags & MDB_WRITEMAP));
+    PyDict_SetItemString(dct, "meminit", py_bool(!(flags & MDB_NOMEMINIT)));
+    return dct;
+}
+
+/**
+ * Environment.max_key_size() -> int
+ */
+static PyObject *
+env_max_key_size(EnvObject *self)
+{
+    if(! self->valid) {
+        return err_invalid();
+    }
+    int key_size = mdb_env_get_maxkeysize(self->env);
+    return PyLong_FromLongLong(key_size);
+}
+
+/**
+ * Environment.max_key_size() -> int
+ */
+static PyObject *
+env_max_readers(EnvObject *self)
+{
+    if(! self->valid) {
+        return err_invalid();
+    }
+    unsigned int readers;
+    int rc = mdb_env_get_maxreaders(self->env, &readers);
+    if(rc) {
+        return err_set("mdb_env_get_maxreaders", rc);
+    }
+    return PyLong_FromLongLong(readers);
+}
+
 /**
  * Environment.open_db() -> handle
  */
@@ -1410,7 +1567,7 @@ env_open_db(EnvObject *self, PyObject *args, PyObject *kwds)
         {ARG_BOOL, CREATE_S, OFFSET(env_open_db, create)},
     };
 
-    if(parse_args(1, SPECSIZE(), argspec, args, kwds, &arg)) {
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
         return NULL;
     }
 
@@ -1571,6 +1728,9 @@ static struct PyMethodDef env_methods[] = {
     {"copy", (PyCFunction)env_copy, METH_VARARGS},
     {"copyfd", (PyCFunction)env_copyfd, METH_VARARGS},
     {"info", (PyCFunction)env_info, METH_NOARGS},
+    {"flags", (PyCFunction)env_flags, METH_NOARGS},
+    {"max_key_size", (PyCFunction)env_max_key_size, METH_NOARGS},
+    {"max_readers", (PyCFunction)env_max_readers, METH_NOARGS},
     {"open_db", (PyCFunction)env_open_db, METH_VARARGS|METH_KEYWORDS},
     {"path", (PyCFunction)env_path, METH_NOARGS},
     {"stat", (PyCFunction)env_stat, METH_NOARGS},
@@ -1716,9 +1876,7 @@ _cursor_get(CursorObject *self, enum MDB_cursor_op op)
     if(_cursor_get_c(self, op)) {
         return NULL;
     }
-    PyObject *res = self->positioned ? Py_True : Py_False;
-    Py_INCREF(res);
-    return res;
+    return py_bool(self->positioned);
 }
 
 /**
@@ -1730,7 +1888,7 @@ cursor_delete(CursorObject *self)
     if(! self->valid) {
         return err_invalid();
     }
-    PyObject *ret = Py_False;
+    PyObject *res = Py_False;
     if(self->positioned) {
         DEBUG("deleting key '%.*s'",
               (int) self->key.mv_size,
@@ -1741,11 +1899,11 @@ cursor_delete(CursorObject *self)
         if(rc) {
             return err_set("mdb_cursor_del", rc);
         }
-        ret = Py_True;
+        res = Py_True;
         _cursor_get_c(self, MDB_GET_CURRENT);
     }
-    Py_INCREF(ret);
-    return ret;
+    Py_INCREF(res);
+    return res;
 }
 
 /**
@@ -2402,19 +2560,18 @@ trans_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 static PyObject *
 trans_abort(TransObject *self)
 {
-    if(! self->valid) {
-        return err_invalid();
-    }
-    DEBUG("aborting")
-    INVALIDATE(self)
+    if(self->valid) {
+        DEBUG("aborting")
+        INVALIDATE(self)
 #ifdef HAVE_MEMSINK
-    ms_notify((PyObject *) self, &self->sink_head);
+        ms_notify((PyObject *) self, &self->sink_head);
 #endif
-    DROP_GIL
-    mdb_txn_abort(self->txn);
-    LOCK_GIL
-    self->txn = NULL;
-    self->valid = 0;
+        DROP_GIL
+        mdb_txn_abort(self->txn);
+        LOCK_GIL
+        self->txn = NULL;
+        self->valid = 0;
+    }
     Py_RETURN_NONE;
 }
 
